import torch
from albumentations import Compose, Normalize
from albumentations.pytorch import ToTensorV2
from PIL import Image
import numpy as np
import io
from torch import nn
from aiia import AIIABase

class Upscaler(nn.Module):
    """
    Transforms the base model's final feature map using a transposed convolution.
    The base model produces a feature map of size 512x512.
    This layer upsamples by a factor of 2 (yielding 1024x1024) and maps the hidden features
    to the output channels using a single ConvTranspose2d layer.
    """
    def __init__(self, base_model: AIIABase):
        super(Upscaler, self).__init__()
        self.base_model = base_model
        # Instead of adding separate upsampling and convolutional layers, we use a ConvTranspose2d layer.
        self.last_transform = nn.ConvTranspose2d(
            in_channels=base_model.config.hidden_size,
            out_channels=base_model.config.num_channels,
            kernel_size=base_model.config.kernel_size,
            stride=2,
            padding=1,
            output_padding=1
        )
        
    def forward(self, x):
        features = self.base_model(x)
        return self.last_transform(features)

class ImageUpscaler:
    def __init__(self, model_path: str, device: str = 'cuda' if torch.cuda.is_available() else 'cpu'):
        """
        Initialize the ImageUpscaler with the trained model.
        
        Args:
            model_path (str): Path to the trained model directory.
            device (str): Device to run inference on ('cuda' or 'cpu').
        """
        self.device = torch.device(device)
        self.model = self.load_model(model_path)
        self.model.eval()  # Set the model to evaluation mode
        
        # Define preprocessing transformations
        self.preprocess = Compose([
            Normalize(mean=[0.5, 0.5, 0.5], std=[0.5, 0.5, 0.5]),
            ToTensorV2()
        ])
    
    def load_model(self, model_path: str):
        """
        Load the trained model from the specified path.
        
        Args:
            model_path (str): Path to the saved model.
        
        Returns:
            nn.Module: Loaded PyTorch model.
        """
        # Load the base model and wrap it with Upscaler
        base_model = AIIABase.load(model_path)
        model = Upscaler(base_model)
        
        # Move the model to the appropriate device
        return model.to(self.device)
    
    def preprocess_image(self, image: Image.Image):
        """
        Preprocess the input image for inference.
        
        Args:
            image (PIL.Image.Image): Input image in PIL format.
        
        Returns:
            torch.Tensor: Preprocessed image tensor.
        """
        # Convert PIL image to numpy array
        image_array = np.array(image)
        
        # Apply preprocessing transformations
        augmented = self.preprocess(image=image_array)
        
        # Add batch dimension and move to device
        return augmented['image'].unsqueeze(0).to(self.device)
    
    def postprocess_image(self, output_tensor: torch.Tensor):
        """
        Postprocess the output tensor to convert it back to an image.
        
        Args:
            output_tensor (torch.Tensor): Model output tensor.
        
        Returns:
            PIL.Image.Image: Upscaled image in PIL format.
        """
        # Remove batch dimension and move to CPU
        output_tensor = output_tensor.squeeze(0).cpu()
        
        # Denormalize and convert to numpy array
        output_array = (output_tensor * 0.5 + 0.5).clamp(0, 1).numpy()
        
        # Convert from CHW (Channels-Height-Width) to HWC (Height-Width-Channels) format
        output_array = (output_array.transpose(1, 2, 0) * 255).astype(np.uint8)
        
        # Convert numpy array back to PIL image
        return Image.fromarray(output_array)
    
    def upscale_image(self, input_image_path: str):
        """
        Perform upscaling on an input image.
        
        Args:
            input_image_path (str): Path to the input low-resolution image.
        
        Returns:
            PIL.Image.Image: Upscaled high-resolution image.
        """
        # Load and preprocess the input image
        input_image = Image.open(input_image_path).convert('RGB')
        preprocessed_image = self.preprocess_image(input_image)
        
        # Perform inference with the model
        with torch.no_grad():
            with torch.cuda.amp.autocast(device_type="cuda"):
                output_tensor = self.model(preprocessed_image)
        
        # Postprocess and return the upscaled image
        return self.postprocess_image(output_tensor)

# Example usage:
# upscaler = ImageUpscaler(model_path="/path/to/best_model")
# upscaled_image = upscaler.upscale_image("/path/to/low_res_image.jpg")
# upscaled_image.save("/path/to/upscaled_image.jpg")
